"""
Overview:
    Function operations for nested structure.
"""
__all__ = [
    'nested_map'
]


def nested_map(f, s):
    """
    Overview:
        Map the nested structure with a function.

    Arguments:
        - f: The given function.
        - s: Nested structure.

    Returns:
        - result: Mapped nested structure.

    Examples::
        >>> from hbutils.collection import nested_map
        >>> nested_map(lambda x: x + 1, [
        ...     2, 3, (4, {'x': 2, 'y': 4}),
        ...     {'a': 3, 'b': (4, 5)},
        ... ])
        [3, 4, (5, {'x': 3, 'y': 5}), {'a': 4, 'b': (5, 6)}]
    """
    if isinstance(s, dict):
        return type(s)({k: nested_map(f, v) for k, v in s.items()})
    elif isinstance(s, (list, tuple)):
        return type(s)(nested_map(f, v) for v in s)
    else:
        return f(s)



"""
Overview:
    Functions to deal with encoding binary data easily.
"""
from typing import Optional

import chardet

_DEFAULT_ENCODING = 'utf-8'
_ENCODING_LIST = ['utf-8', 'gbk', 'gb2312', 'gb18030', 'big5']  # common encodings for chinese

__all__ = [
    'auto_decode'
]


def auto_decode(data: bytes, encoding: Optional[str] = None) -> str:
    r"""
    Overview:
        Auto decode binary data to string, the encoding mode will be automatically detected.

    Arguments:
        - data (:obj:`bytes`): Original binary data to be decoded.
        - encoding (:obj:`Optional[str]`): Encoding mode to be used, default is ``None`` which \
            means this function need to automatically detect the encoding.

    Returns:
        - str (:obj:`str`): Decoded string.

    Examples::

        >>> auto_decode(b'kdsfjldsjflkdsmgds')  # 'kdsfjldsjflkdsmgds'
        >>> auto_decode(b'\xd0\x94\xd0\xbe\xd0\xb1\xd1\x80\xd1\x8b\xd0\xb9 \xd0'
        ...             b'\xb2\xd0\xb5\xd1\x87\xd0\xb5\xd1\x80')  # "Добрый вечер"
        >>> auto_decode(b'\xa4\xb3\xa4\xf3\xa4\xd0\xa4\xf3\xa4\xcf')  # "こんばんは"
        >>> auto_decode(b'\xcd\xed\xc9\xcf\xba\xc3')  # "晚上好"
    """
    if encoding:
        return data.decode(encoding)
    else:
        auto_encoding = chardet.detect(data)['encoding']
        if auto_encoding and auto_encoding not in _ENCODING_LIST:
            _list = _ENCODING_LIST + [auto_encoding]
        else:
            _list = _ENCODING_LIST

        last_err = None
        for enc in _list:
            try:
                return data.decode(encoding=enc)
            except UnicodeDecodeError as err:
                last_err = err

        raise last_err



"""
Overview:
    Structural operations.
"""
__all__ = [
    'sq_flatten',
]


def _g_sq_flatten(s):
    if isinstance(s, (list, tuple)):
        for item in s:
            yield from _g_sq_flatten(item)
    else:
        yield s


def sq_flatten(s):
    """
    Overview:
        Sequence flatten.
    Arguments:
        - s: The given sequence.

    Returns:
        - flatted: Flatted sequence.

    Examples::
        >>> from hbutils.collection import sq_flatten
        >>> sq_flatten([1, 2, [3, 4], [5, [6, 7], (8, 9, 10)], 11])
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    """
    return list(_g_sq_flatten(s))
